# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Saturation Environment
=======================
"""
import os
import random
from typing import Dict, List, Tuple

import numpy as np
from gym import Env, spaces

from gym_saturation.clause_space import ClauseSpace
from gym_saturation.grammar import Clause
from gym_saturation.logic_ops.factoring import all_possible_factors
from gym_saturation.logic_ops.paramodulation import all_paramodulants_from_list
from gym_saturation.logic_ops.reflexivity_resolution import (
    all_possible_reflexivity_resolvents,
)
from gym_saturation.logic_ops.resolution import all_possible_resolvents
from gym_saturation.logic_ops.utils import (
    clause_in_a_list,
    is_tautology,
    reduce_to_proof,
    reindex_variables,
)
from gym_saturation.parsing.json_grammar import clause_to_dict
from gym_saturation.parsing.tptp_parser import TPTPParser

STATE_DIFF_UPDATED = "state_diff_updated"
MAX_CLAUSES = 100000


class SaturationEnv(Env):
    """
    saturation algorithm defined in a Reiforcement Learning friendly way

    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> tptp_folder = files("gym_saturation").joinpath(
    ...     os.path.join("resources", "TPTP-mock")
    ... )
    >>> from glob import glob
    >>> problem_list = sorted(glob(
    ...     os.path.join(tptp_folder, "Problems", "*", "*1-1.p")
    ... ))
    >>> env = SaturationEnv(problem_list, 1)
    >>> env.seed(0)
    0
    >>> env.reset()
    Traceback (most recent call last):
     ...
    ValueError: Too many clauses: 4
    consider increasing `max_clauses` parameter of `__init__`
    >>> env = SaturationEnv(problem_list)
    >>> env.seed(0)
    0
    >>> len(env.reset()["real_obs"])
    4

    one can look at the current state in TPTP format

    >>> print(env.render())
    cnf(this_is_a_test_case_1, hypothesis, this_is_a_test_case(test_constant)).
    cnf(this_is_a_test_case_2, hypothesis, ~this_is_a_test_case(test_constant)).
    cnf(test_axiom, hypothesis, test_constant = X0).
    cnf(test_axiom_2, hypothesis, ~test_constant = 0).

    ``ansi`` mode returns a JSON representation of the state
    it should be more easily parsable than TPTP, although less human-friendly

    >>> env.render("ansi")[:73]
    "[{'class': 'Clause', 'literals': [{'class': 'Literal', 'negated': False, "

    other modes are not implemented yet

    >>> env.render(mode="rgb_array")
    Traceback (most recent call last):
     ...
    NotImplementedError

    the test theorem can be proved in three steps

    >>> next_state, reward, done, info = env.step(0)

    ``info`` dict contains the state diff, for example

    >>> info["state_diff_updated"][0]["processed"]
    True

    repeating actions is not allowed

    >>> env.step(0)
    Traceback (most recent call last):
     ...
    ValueError: action 0 is not valid

    there is no reward until the end of an episode

    >>> env.step(1)[1:3]
    (0.0, False)

    if a proof is found, then reward is ``+1``

    >>> env.step(4)[1:3]
    (1.0, True)

    TSTP proof is now available (one can add ``include`` directive before it
    for validation purposes)

    >>> print(TPTPParser().parse(env.tstp_proof, "")[0])  # doctest: +ELLIPSIS
    cnf(..., hypothesis, $false(), inference(resolution, [], [this_is_a_test_case_1, this_is_a_test_case_2])).
    """

    def __init__(
        self,
        problem_list: List[str],
        max_clauses: int = MAX_CLAUSES,
    ):
        super().__init__()
        self.problem_list = problem_list
        self._state: List[Clause] = []
        self.action_space = spaces.Discrete(max_clauses)
        self.observation_space = spaces.Dict(
            {
                "action_mask": spaces.Box(0, 1, shape=(max_clauses,)),
                "real_obs": ClauseSpace(),
            }
        )

    def _init_clauses(self):
        problem = random.choice(self.problem_list)
        tptp_folder = os.path.join(os.path.dirname(problem), "..", "..")
        with open(problem, "r", encoding="utf-8") as problem_file:
            problem_text = problem_file.read()
        clauses = TPTPParser().parse(problem_text, tptp_folder)
        for clause in clauses:
            clause.birth_step = 0
            clause.inference_parents = []
            clause.inference_rule = None
            clause.processed = False
        return clauses

    def reset(self) -> dict:
        self._state = reindex_variables(self._init_clauses(), "X")
        return self.state

    def _add_to_state(self, new_clauses: List[Clause]) -> None:
        birth_step = 1 + max(
            [getattr(clause, "birth_step", 0) for clause in self._state]
        )
        for clause in new_clauses:
            if not is_tautology(clause):
                if not clause_in_a_list(clause, self._state):
                    clause.birth_step = birth_step
                    clause.processed = False
                    self._state.append(clause)

    def _do_deductions(self, action: int) -> Dict[int, Clause]:
        state_len_before = len(self._state)
        given_clause = self._state[action]
        if not given_clause.processed:
            unprocessed_clauses = [
                clause for clause in self._state if clause.processed
            ]
            self._add_to_state(
                all_possible_resolvents(
                    unprocessed_clauses,
                    given_clause,
                )
            )
            self._add_to_state(
                all_paramodulants_from_list(
                    unprocessed_clauses,
                    given_clause,
                )
            )
            self._add_to_state(
                all_possible_factors(
                    given_clause,
                )
            )
            self._add_to_state(
                all_possible_reflexivity_resolvents(
                    given_clause,
                )
            )
        self._state[action].processed = True
        return dict(
            [
                (i + state_len_before, clause_to_dict(clause))
                for i, clause in enumerate(self._state[state_len_before:])
            ]
            + [(action, clause_to_dict(self._state[action]))]
        )

    def step(self, action: int) -> Tuple[dict, float, bool, dict]:
        if self._state[action].processed:
            raise ValueError(f"action {action} is not valid")
        if self._state[action].literals == []:
            self._state[action].processed = True
            return (
                self.state,
                1.0,
                True,
                {
                    STATE_DIFF_UPDATED: {
                        action: clause_to_dict(self._state[action])
                    }
                },
            )
        updated = self._do_deductions(action)
        if min(
            [
                False if clause.processed is None else clause.processed
                for clause in self._state
            ]
        ):
            return self.state, 0.0, True, {STATE_DIFF_UPDATED: updated}
        return (
            self.state,
            0.0,
            False,
            {STATE_DIFF_UPDATED: updated},
        )

    # pylint: disable=inconsistent-return-statements
    def render(self, mode="human"):
        if mode == "ansi":
            return str(self.state["real_obs"])
        if mode == "human":
            return "\n".join(map(str, self._state))
        super().render(mode=mode)

    @property
    def state(self) -> dict:
        """
        :returns: environment state in Python ``dict`` format
        """
        if len(self._state) > self.action_space.n:
            raise ValueError(
                f"Too many clauses: {len(self._state)}\n"
                "consider increasing `max_clauses` parameter of `__init__`"
            )
        return {
            "real_obs": clause_to_dict(self._state),
            "action_mask": (
                np.array(
                    [
                        0.0 if clause.processed else 1.0
                        for clause in self._state
                    ]
                    + self.action_space.n * [0.0],
                    np.float32,
                )
            )[: self.action_space.n],
        }

    def seed(self, seed=None):
        random.seed(seed)
        return seed

    @property
    def tstp_proof(self) -> str:
        """
        :returns: TSTP proof (if found; raises error otherwise)
        """
        return "\n".join(
            reversed(
                [
                    str(clause)
                    for clause in reduce_to_proof(self._state)
                    if clause.inference_rule is not None
                ]
            )
        )



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Logic Operations Utils
=======================
"""
from itertools import chain
from typing import List, Union

from gym_saturation.grammar import (
    Clause,
    Function,
    Predicate,
    Proposition,
    Term,
    Variable,
)
from gym_saturation.logic_ops.substitution import Substitution
from gym_saturation.utils import deduplicate, pickle_copy


def is_subproposition(one: Proposition, two: Proposition) -> bool:
    """
    check whether proposition ``one`` is part of a proposition ``two``

    :param one: presumably shorter proposition which is probably a part
    :param two: presumably longer proposition which is probably the whole
    :returns: whether ``one`` is a part of ``two``
    """
    if isinstance(two, Variable):
        return one == two
    if isinstance(two, (Function, Predicate)):
        for argument in two.arguments:
            if is_subproposition(one, argument):
                return True
    return False


def get_variable_list(clause: Union[Clause, Proposition]) -> List[Variable]:
    """
    find all variables present in a clause

    >>> from gym_saturation.grammar import Literal
    >>> get_variable_list(Clause([Literal(False, Predicate("this_is_a_test_case", [Function("f", [Variable("X"), Variable("X")])]))]))
    [Variable(name='X'), Variable(name='X')]

    :param clause: a clause
    :returns: a list (with repetitions) of variables from there clause
    """
    variable_list = []
    if isinstance(clause, Clause):
        for literal in clause.literals:
            for term in literal.atom.arguments:
                variable_list.extend(get_variable_list(term))
    elif isinstance(clause, Function):
        for term in clause.arguments:
            variable_list.extend(get_variable_list(term))
    elif isinstance(clause, Variable):
        variable_list.append(clause)
    return variable_list


def reindex_variables(clauses: List[Clause], prefix: str) -> List[Clause]:
    """
    rename variables in a list of clauses so that each clause has its unique
    set of variables

    :param clauses: a list of clauses
    :param prefix: new variables will be named ``prefix[order_num]``
    :returns: the list of clauses with renamed variables
    """
    variable_count = 0
    new_clauses = []
    for clause in clauses:
        new_clause = clause
        variable_list = deduplicate(get_variable_list(clause))
        new_variables_count = len(variable_list)
        for i in range(new_variables_count):
            new_clause = Substitution(
                variable_list[i], Variable(f"{prefix}{i + variable_count}")
            ).substitute_in_clause(new_clause)
        variable_count += new_variables_count
        new_clauses.append(new_clause)
    return new_clauses


def is_tautology(clause: Clause) -> bool:
    """
    check whether there are two literals, one negated and the other not, with
    the same atom

    >>> from gym_saturation.grammar import Literal
    >>> is_tautology(Clause([Literal(False, Predicate("this_is_a_test_case", []))]))
    False
    >>> is_tautology(Clause([Literal(False, Predicate("this_is_a_test_case", [])), Literal(True, Predicate("this_is_a_test_case", []))]))
    True
    >>> is_tautology(Clause([Literal(False, Predicate("=", [Variable("X"), Variable("X")]))], label="this_is_a_test_case"))
    True

    :param clause: a clause to check
    :returns: whether the clause is a primitive tautology or not
    """
    for i, literal in enumerate(clause.literals):
        for j, another_literal in enumerate(clause.literals):
            if (
                i != j
                and literal.negated != another_literal.negated
                and literal.atom == another_literal.atom
            ):
                return True
        if literal.atom.name == "=" and (
            literal.atom.arguments[0] == literal.atom.arguments[1]
        ):
            return True
    return False


def clause_length(clause: Clause) -> int:
    """
    total length of arguments of each predicate.
    Negation adds one to each literal

    :param clause: a clause
    :return: sctructural length of a clause

    >>> from gym_saturation.grammar import Literal
    >>> clause_length(Clause([Literal(True, Predicate("p", [Function("this_is_a_test_case", [])]))]))
    3
    """
    length = 0
    for literal in clause.literals:
        if literal.negated:
            length += 1
        for term in literal.atom.arguments:
            length += proposition_length(term)
        length += 1
    return length


def proposition_length(proposition: Proposition) -> int:
    """
    total number of functional, predicate and variable symbols

    :param proposition: a function, a predicate or a variable
    :return: sctructural length of a proposition

    >>> proposition_length(Predicate("p", [Function("f", [Variable("X")])]))
    3
    """
    length = 0
    if isinstance(proposition, Variable):
        return 1
    for subterm in proposition.arguments:
        length += proposition_length(subterm)
    return 1 + length


def clause_in_a_list(clause: Clause, clauses: List[Clause]) -> bool:
    """

    >>> clause_in_a_list(Clause([], label="one"), [Clause([], label="two")])
    True

    :param clause: some clause
    :param clauses: a list of clauses
    :returns: whether in the list there is a clause with a literals set to a
        given clause
    """
    for a_clause in clauses:
        if a_clause.literals == clause.literals:
            return True
    return False


class NoSubtermFound(Exception):
    """sometimes a subterm index is larger than term length"""


def subterm_by_index(atom: Proposition, index: int) -> Term:
    """
    extract a subterm using depth-first search through the tree of logical
    operations

    >>> atom = Predicate("this_is_a_test_case", [Function("f", [Variable("X")]), Function("g", [Variable("Y")])])
    >>> subterm_by_index(atom, 0)
    Traceback (most recent call last):
     ...
    ValueError: subterm with index 0 exists only for terms, but got: Predicate(name='this_is_a_test_case', arguments=[Function(name='f', arguments=[Variable(name='X')]), Function(name='g', arguments=[Variable(name='Y')])])
    >>> subterm_by_index(atom, 1) == atom.arguments[0]
    True
    >>> subterm_by_index(atom, 2) == atom.arguments[0].arguments[0]
    True
    >>> subterm_by_index(atom, 4) == atom.arguments[1].arguments[0]
    True

    :param atom: a predicate or a term
    :param index: an index of a desired subterm
    :returns: a subterm
    """
    if index == 0:
        if isinstance(atom, (Function, Variable)):
            return atom
        raise ValueError(
            f"subterm with index 0 exists only for terms, but got: {atom}"
        )
    subterm_length = 1
    if not isinstance(atom, Variable):
        for argument in atom.arguments:
            try:
                return pickle_copy(
                    subterm_by_index(argument, index - subterm_length)
                )
            except NoSubtermFound as error:
                subterm_length += error.args[0]
    raise NoSubtermFound(subterm_length)


class CantReplaceTheWholeTerm(Exception):
    """an exception raised when trying to replace a subterm with index 0"""


class TermSelfReplace(Exception):
    """an exception raised when trying to replace a subterm with itself"""


def _replace_if_not_the_same(
    atom: Proposition, index: int, term: Term
) -> None:
    if not isinstance(atom, Variable):
        if atom.arguments[index] == term:
            raise TermSelfReplace
        atom.arguments[index] = term


def replace_subterm_by_index(
    atom: Proposition, index: int, term: Term
) -> None:
    """
    replace a subterm with a given index (depth-first search) by a new term
    replacement always happens inplace!

    >>> atom = Predicate("this_is_a_test_case", [Function("f", [Variable("X")]), Function("g", [Variable("Y")])])
    >>> replace_subterm_by_index(atom, 0, Variable("Z"))
    Traceback (most recent call last):
     ...
    gym_saturation.logic_ops.utils.NoSubtermFound: 5
    >>> replace_subterm_by_index(atom, 4, Function("h", [Variable("Z")]))
    >>> atom
    Predicate(name='this_is_a_test_case', arguments=[Function(name='f', arguments=[Variable(name='X')]), Function(name='g', arguments=[Function(name='h', arguments=[Variable(name='Z')])])])
    >>> replace_subterm_by_index(Predicate("this_is_a_test_case", [Variable("X")]), 1, Variable("X"))
    Traceback (most recent call last):
     ...
    gym_saturation.logic_ops.utils.TermSelfReplace

    :param atom: a predicate or a term
    :param index: an index of a subterm to replace, must be greater than 0
    :param term: replacement term for a given index
    :returns:
    """
    subterm_length = 1
    if not isinstance(atom, Variable):
        for i, argument in enumerate(atom.arguments):
            if index == subterm_length:
                _replace_if_not_the_same(atom, i, term)
                return
            try:
                replace_subterm_by_index(
                    argument, index - subterm_length, term
                )
                return
            except NoSubtermFound as error:
                subterm_length += error.args[0]
    raise NoSubtermFound(subterm_length)


def _flat_list(list_of_lists: List[List[str]]) -> List[str]:
    return list(reversed(sorted(list(set(chain(*list_of_lists))))))


def reduce_to_proof(clauses: List[Clause]) -> List[Clause]:
    """
    leave only clauses belonging to the refutational proof

    >>> reduce_to_proof([Clause([])])
    Traceback (most recent call last):
     ...
    ValueError: wrong refutation proof
    >>> state = [Clause([], label="one", processed=True)]
    >>> reduce_to_proof(state) == state
    True

    :param clauses: a list of clauses with labels and inference records
    :returns: the reduced list of clauses
    """
    state_dict = {clause.label: clause for clause in clauses}
    empty_clauses = [
        clause
        for label, clause in state_dict.items()
        if clause.literals == [] and clause.processed
    ]
    if len(empty_clauses) == 1:
        if empty_clauses[0].label is not None:
            reduced = []
            new_reduced = [empty_clauses[0]]
            while len(new_reduced) > 0:
                reduced += [
                    clause for clause in new_reduced if clause not in reduced
                ]
                new_reduced = [
                    state_dict[label]
                    for label in _flat_list(
                        [
                            (
                                []
                                if clause.inference_parents is None
                                else clause.inference_parents
                            )
                            for clause in new_reduced
                        ]
                    )
                ]
            return reduced
    raise ValueError("wrong refutation proof")



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Agent Testing
==============

This module is an example of testing your own trained agent.
"""
import random
import sys
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from logging import Logger
from operator import itemgetter
from typing import Any, Dict, List, Optional

import gym
from gym.wrappers import TimeLimit

from gym_saturation.envs import SaturationEnv
from gym_saturation.envs.saturation_env import STATE_DIFF_UPDATED
from gym_saturation.logic_ops.utils import clause_length
from gym_saturation.parsing.json_grammar import dict_to_clause


class BaseAgent:
    """a basic RL agent"""

    _state = None

    @property
    def state(self):
        """agent can have its inner state"""
        return self._state

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:
        """
        :param observation: an observation returned by the environment after
            the latest action (or it's initial state)
        :param reward: reward from the previous step
            (use zero for the first step)
        :param info: info dictionary returned by the environment
        :returns: an action
        """


@dataclass
class Transition:
    """
    an object describing environment's and agent's states
    before the agent's action, the action itself and its results
    """

    env_state: dict
    agent_state: Any
    action: int
    observation: dict
    reward: float
    done: bool
    info: Dict[str, Any]


class SizeAgent(BaseAgent):
    """
    .. _size_agent:

    agent which selects the shortest clause
    """

    def __init__(self):
        self._state: Dict[int, float] = {}

    def update_state(self, info: Dict[str, Any]) -> None:
        """
        update the state of the agent according with the transition

        :param info: an info dict (parm of environment response)
        :returns:
        """
        self._state.update(
            {
                index: clause_length(dict_to_clause(clause))
                for index, clause in info[STATE_DIFF_UPDATED].items()
                if not clause["processed"]
            }
        )
        for index, clause in info[STATE_DIFF_UPDATED].items():
            if clause["processed"]:
                self._state.pop(index)

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:
        self.update_state(info)
        return min(
            self._state.items(),
            key=itemgetter(1),
        )[0]


class AgeAgent(BaseAgent):
    """
    .. _age_agent:

    agent which selects the oldest clause
    """

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:
        return min(
            [
                i
                for i, clause in enumerate(observation["real_obs"])
                if not clause["processed"]
            ]
        )


class SizeAgeAgent(BaseAgent):
    """
    .. _size_age_agent:

    agent which takes several times the smallest clause and then several
    times the oldest
    """

    def __init__(self, size_steps: int, age_steps: int):
        self.size_steps = size_steps
        self.age_steps = age_steps
        self._step_count = 0
        self._use_size = True
        self._size_agent = SizeAgent()
        self._age_agent = AgeAgent()

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:
        self._step_count += 1
        if self._use_size:
            if self._step_count >= self.size_steps:
                self._step_count = 0
                self._use_size = False
            return self._size_agent.get_action(observation, reward, info)
        if self._step_count >= self.age_steps:
            self._step_count = 0
            self._use_size = True
        self._size_agent.update_state(info)
        return self._age_agent.get_action(observation, reward, info)


class RandomAgent(BaseAgent):
    """agent which selects clauses randomly"""

    def get_action(
        self,
        observation: dict,
        reward: float,
        info: Dict[str, Any],
    ) -> int:
        return random.choice(
            [
                i
                for i, clause in enumerate(observation["real_obs"])
                if not clause["processed"]
            ]
        )


def episode(
    env: SaturationEnv, agent: BaseAgent, logger: Optional[Logger] = None
) -> Transition:
    """
    tries to solve the problem and logs the clauses

    >>> import os
    >>> import shutil
    >>> test_agent_output = "test_agent_output"
    >>> shutil.rmtree(test_agent_output, ignore_errors=True)
    >>> os.mkdir(test_agent_output)
    >>> import sys
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> from glob import glob
    >>> problem_list = sorted(glob(os.path.join(
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join(
    ...         "resources", "TPTP-mock", "Problems", "TST"
    ...     ))
    ... , "*-*.p")))
    >>> random.seed(0)
    >>> agents = [SizeAgeAgent(2, 1), SizeAgeAgent(1, 2), RandomAgent()]
    >>> for i in range(3):  # doctest: +ELLIPSIS
    ...     env = gym.make(
    ...         "gym_saturation:saturation-v0",
    ...         problem_list=[problem_list[i]],
    ...     )
    ...     env._max_episode_steps = 5
    ...     agent_testing_report(env, agents[i])
    Proof of length 1 found in 4 steps:
    cnf(..., hypothesis, $false, inference(resolution, [], [this_is_a_test_case_1, this_is_a_test_case_2])).
    Step limit reached
    Proof of length 2 found in 5 steps:
    cnf(..., ..., ~man(socrates), inference(resolution, [], [q, p_imp_q])).
    cnf(..., hypothesis, $false, inference(resolution, [], [p, ...])).

    :param env: a `gym_saturation` environment
    :param agent: an initialized agent. Must have `get_action` method
    :param problem_filename: the name of a problem file
    :param logger: where to log the episode memory;
        if ``None`` then nothing is logged
    :returns: the last transition
    """
    env_state, reward, done = env.reset(), 0.0, False
    info: Dict[str, Any] = {
        STATE_DIFF_UPDATED: dict(enumerate(env_state["real_obs"]))
    }
    while not done:
        action = agent.get_action(env_state, reward, info)
        observation, reward, done, info = env.step(action)
        transition = Transition(
            env_state,
            agent.state,
            action,
            observation,
            reward,
            done,
            info,
        )
        if logger is not None:
            logger.info(transition)
        env_state = observation
    return transition


def parse_args(args: Optional[List[str]] = None) -> Namespace:
    """
    >>> parse_args([
    ...     "--problem_filename", "test",
    ...     "--step_limit", "1"
    ... ])
    Namespace(problem_filename='test', step_limit=1)

    :param args: a list of string arguments
        (for testing and use in a non script scenario)
    :returns: arguments namespace for the script
    """
    argument_parser = ArgumentParser()
    argument_parser.add_argument("--problem_filename", type=str, required=True)
    argument_parser.add_argument("--step_limit", type=int, required=True)
    parsed_args = argument_parser.parse_args(args)
    return parsed_args


def agent_testing_report(env: SaturationEnv, agent: BaseAgent) -> None:
    """
        print a report after testing an agent in an environment

    :param env: an environment
    :param agent: an agent
    :returns:
    """
    last_transition = episode(env, agent)
    if last_transition.reward == 1.0:
        a_proof = env.tstp_proof
        proof_length = len(a_proof.split("\n"))
        step_count = getattr(env, "_elapsed_steps")
        print(
            f"Proof of length {proof_length} found " f"in {step_count} steps:"
        )
        print(a_proof)
    else:
        print("Step limit reached")


def main(args: Optional[List[str]] = None) -> None:
    """
    the main function for this module

    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> import os
    >>> problem_filename = os.path.join(
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join(
    ...         "resources", "TPTP-mock", "Problems", "TST", "TST001-1.p"
    ...     ))
    ... )
    >>> main([
    ...     "--problem_filename", problem_filename,
    ...     "--step_limit", "3"
    ... ]) # doctest: +ELLIPSIS
    Problem file: ...TST001-1.p
    Proof of length 1 found in 3 steps:
    cnf(..., hypothesis, $false, inference(resolution, [], [this_is_a_test_case_1, this_is_a_test_case_2])).

    """
    sys.setrecursionlimit(10000)
    arguments = parse_args(args)
    environment = TimeLimit(
        gym.make(
            "gym_saturation:saturation-v0",
            problem_list=[arguments.problem_filename],
        ),
        arguments.step_limit,
    )
    print(f"Problem file: {arguments.problem_filename}")
    agent_testing_report(environment, SizeAgeAgent(5, 1))


if __name__ == "__main__":
    main()



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
JSONEncoder for the grammar
============================
"""

import json
from typing import Dict, List, Union

from gym_saturation import grammar


class ClauseJSONEncoder(json.JSONEncoder):
    """
    by default ``Clause`` is not serializable type. This class is a standard
    way to make it such. It's coupled with a function ``dict_to_clause`` to
    restore serialised objects back. Potentially that can be done in any
    language, not only Python.

    >>> this_is_a_test_case = grammar.Clause([grammar.Literal(False, grammar.Predicate("=", [grammar.Function("f", [grammar.Variable("X"), grammar.Function("g", [grammar.Variable("Y")]), grammar.Function("h", [grammar.Variable("Z"), grammar.Function("c1", [])])]), grammar.Function("f", [grammar.Variable("X"), grammar.Variable("Y"), grammar.Function("c2", [])])])), grammar.Literal(True, grammar.Predicate("better", [grammar.Variable("X"), grammar.Function("g", [grammar.Variable("Y")])]))], label="some", birth_step=0, inference_rule="resolution", inference_parents=["one", "two"], processed=False)
    >>> # this is a standard way to serialise a ``Clause``
    >>> serialised_clause = clause_to_dict(this_is_a_test_case)
    >>> # and this is how to deserialise
    >>> deserialised_clause = dict_to_clause(serialised_clause)
    >>> # sanity check
    >>> print(this_is_a_test_case == deserialised_clause)
    True
    >>> dict_to_clause({"class": "not a Clause!"})
    Traceback (most recent call last):
     ...
    ValueError: json is not a Clause: {'class': 'not a Clause!'}
    """

    @staticmethod
    def _variable_to_dict(variable: grammar.Variable) -> Dict[str, str]:
        return {"class": "Variable", "name": variable.name}

    def _function_to_dict(
        self, function: grammar.Function
    ) -> Dict[str, Union[str, list]]:
        arguments: List[Union[str, dict]] = []
        for argument in function.arguments:
            if isinstance(argument, grammar.Variable):
                arguments.append(self._variable_to_dict(argument))
            else:
                arguments.append(self._function_to_dict(argument))
        return {
            "class": "Function",
            "name": function.name,
            "arguments": arguments,
        }

    def default(self, o):
        literals = []
        for literal in o.literals:
            arguments = []
            for argument in literal.atom.arguments:
                if isinstance(argument, grammar.Variable):
                    arguments.append(self._variable_to_dict(argument))
                else:
                    arguments.append(self._function_to_dict(argument))
            literals.append(
                {
                    "class": "Literal",
                    "negated": literal.negated,
                    "atom": {
                        "class": "Predicate",
                        "name": literal.atom.name,
                        "arguments": arguments,
                    },
                }
            )
        return {
            "class": "Clause",
            "literals": literals,
            "label": o.label,
            "birth_step": o.birth_step,
            "processed": o.processed,
            "inference_parents": o.inference_parents,
            "inference_rule": o.inference_rule,
        }


def dict_to_clause(json_dict):
    """
    .. _dict_to_clause:

    get ``Clause`` from a Python dictionary
    """
    if json_dict["class"] == "Clause":
        return grammar.Clause(
            [dict_to_clause(literal) for literal in json_dict["literals"]],
            label=json_dict["label"],
            birth_step=json_dict["birth_step"],
            processed=json_dict["processed"],
            inference_parents=json_dict["inference_parents"],
            inference_rule=json_dict["inference_rule"],
        )
    if json_dict["class"] == "Literal":
        return grammar.Literal(
            json_dict["negated"], dict_to_clause(json_dict["atom"])
        )
    if json_dict["class"] == "Predicate":
        return grammar.Predicate(
            json_dict["name"],
            [dict_to_clause(argument) for argument in json_dict["arguments"]],
        )
    if json_dict["class"] == "Function":
        return grammar.Function(
            json_dict["name"],
            [dict_to_clause(argument) for argument in json_dict["arguments"]],
        )
    if json_dict["class"] == "Variable":
        return grammar.Variable(json_dict["name"])
    raise ValueError(f"json is not a Clause: {json_dict}")


def clause_to_dict(clause):
    """convert clause to JSON-like dictionary"""
    return json.loads(json.dumps(clause, cls=ClauseJSONEncoder))



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
CNF Parser
===========
"""
from operator import itemgetter

from lark import Transformer

from gym_saturation.grammar import (
    Clause,
    Function,
    Literal,
    Predicate,
    Variable,
)


class CNFParser(Transformer):
    """
    a parser for ``<cnf_formula>`` from Lark parse tree
    methods are not typed since nobody calls them directly

    >>> import sys
    >>> import os
    >>> if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    ...     from importlib.resources import files
    ... else:
    ...     from importlib_resources import files
    >>> from lark import Lark
    >>> parser = Lark(
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join("resources", "TPTP.lark"))
    ...     .read_text(),
    ...     start="tptp_file"
    ... )
    >>> CNFParser().transform(parser.parse('''
    ...    cnf(test, axiom, f(X, g(Y), h(Z, c1)) = f(X, Y, c2)
    ...    | ~ better(f(X), g(Y)) | $false | this_is_a_test_case).
    ... '''))
    cnf(test, hypothesis, f(X,g(Y),h(Z,c1)) = f(X,Y,c2) | ~better(f(X), g(Y)) | $false() | this_is_a_test_case()).
    """

    def __default_token__(self, token):
        return token.value

    def __default__(self, data, children, meta):
        if len(children) == 1:
            return children[0]
        return children

    @staticmethod
    def _function(children):
        """
        a functional symbol with arguments
        """
        if len(children) > 1:
            return Function(children[0], children[1])
        return Function(children[0], [])

    def fof_defined_plain_formula(self, children):
        """
        <fof_defined_plain_formula> :== <defined_proposition> | <defined_predicate>(<fof_arguments>)
        """
        return self._predicate(children)

    def fof_plain_term(self, children):
        """
        <fof_plain_term>       ::= <constant> | <functor>(<fof_arguments>)
        """
        return self._function(children)

    def fof_defined_term(self, children):
        """
        <fof_defined_term>     ::= <defined_term> | <fof_defined_atomic_term>
        """
        return self._function(children)

    @staticmethod
    def variable(children):
        """
        <variable>             ::= <upper_word>

        a variable (supposed to be universally quantified)
        """
        return Variable(children[0])

    @staticmethod
    def fof_arguments(children):
        """
        <fof_arguments>        ::= <fof_term> | <fof_term>,<fof_arguments>

        a list of arguments, organised in pairs
        """
        result = []
        for item in children:
            if isinstance(item, list):
                result.extend(item)
            else:
                result.append(item)
        return result

    @staticmethod
    def literal(children):
        """
        <literal>              ::= <fof_atomic_formula> | ~ <fof_atomic_formula> | <fof_infix_unary>

        a literal is a possible negated predicate
        """
        if children[0] == "~":
            return Literal(True, children[1])
        if isinstance(children[0], Predicate):
            if children[0].name == "!=":
                children[0].name = "="
                return Literal(True, children[0])
        return Literal(False, children[0])

    @staticmethod
    def _predicate(children):
        """predicates are atomic formulae"""
        if len(children) > 1:
            return Predicate(children[0], children[1])
        return Predicate(children[0], [])

    def fof_plain_atomic_formula(self, children):
        """
        <fof_plain_atomic_formula> :== <proposition> | <predicate>(<fof_arguments>)
        """
        return self._predicate(children)

    @staticmethod
    def fof_defined_infix_formula(children):
        """
        <fof_defined_infix_formula> ::= <fof_term> <defined_infix_pred> <fof_term>

        some predicates are in the infix form, so we translate to them prefix
        """
        return Predicate(children[1], [children[0], children[2]])

    @staticmethod
    def fof_infix_unary(children):
        """
        <fof_infix_unary>      ::= <fof_term> <infix_inequality> <fof_term>

        some predicates are in the infix form, so we translate to them prefix
        """
        return Predicate(children[1], [children[0], children[2]])

    @staticmethod
    def disjunction(children):
        """
        <disjunction>          ::= <literal> | <disjunction> <vline> <literal>

        basic clause structure
        """
        if len(children) == 1:
            return Clause(children)
        literals = []
        for item in [children[0], children[2]]:
            if isinstance(item, Clause):
                literals.extend(item.literals)
            else:
                literals.append(item)
        return Clause(literals)

    @staticmethod
    def _parse_inference_parents(inference_parents):
        if isinstance(inference_parents[0], list):
            clause_inference_parents = list(
                map(
                    itemgetter(0),
                    inference_parents,
                )
            )
        else:
            clause_inference_parents = [inference_parents[0]]
        return clause_inference_parents

    def cnf_annotated(self, children):
        """
        <cnf_annotated>        ::= cnf(<name>,<formula_role>,<cnf_formula> <annotations>).

        annotated CNF formula (clause)
        """
        clause = children[2]
        clause.label = children[0]
        if isinstance(children[3], list):
            for annotation in children[3]:
                if isinstance(annotation, dict):
                    if "inference_record" in annotation:
                        clause.inference_rule = annotation["inference_record"][
                            0
                        ]
                        clause.inference_parents = (
                            self._parse_inference_parents(
                                annotation["inference_record"][1]
                            )
                        )
        return clause

    @staticmethod
    def inference_record(children):
        """
        <inference_record>     :== inference(<inference_rule>,<useful_info>,
        <inference_parents>)
        """
        return {"inference_record": (children[0], children[2])}

    @staticmethod
    def annotations(children):
        """
        <annotations>          ::= ,<source><optional_info> | <null>
        """
        if isinstance(children, list):
            if len(children) == 1:
                return children[0]
        return children

    @staticmethod
    def source(children):
        """
        <source>               ::= <general_term>
        <source>               :== <dag_source> | <internal_source> |
        <external_source> | unknown | [<sources>]
        """
        if isinstance(children, list):
            if len(children) == 1:
                return children[0]
        return children

    @staticmethod
    def dag_source(children):
        """
        <dag_source>           :==  <inference_record> | <name>
        """
        if isinstance(children, list):
            if len(children) == 1:
                return children[0]
        return children



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
TPTP Parser
===========
"""
import os
import sys
from typing import List

from lark import Lark, Token

from gym_saturation.grammar import Clause
from gym_saturation.parsing.cnf_parser import CNFParser

if sys.version_info.major == 3 and sys.version_info.minor >= 9:
    # pylint: disable=no-name-in-module, import-error
    from importlib.resources import files  # type: ignore
else:
    from importlib_resources import files  # pylint: disable=import-error


# pylint: disable=too-few-public-methods
class TPTPParser:
    """
    >>> from gym_saturation.grammar import (Literal, Predicate, Variable,
    ...     Function)
    >>> clause = Clause([Literal(True, Predicate("=", [Function("this_is_a_test_case", [Variable("X")]), Variable("Y")]))], inference_rule="resolution", inference_parents=["one", "two"])
    >>> TPTPParser().parse(str(clause), "") == [clause]
    True
    >>> tptp_parser = TPTPParser()
    >>> tptp_text = (
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join(
    ...         "resources", "TPTP-mock", "Problems", "TST", "TST001-1.p"
    ...     ))
    ...     .read_text()
    ... )
    >>> print("\\n".join(map(str, tptp_parser.parse(
    ...     tptp_text,
    ...     files("gym_saturation")
    ...     .joinpath(os.path.join("resources", "TPTP-mock"))
    ... ))))
    cnf(this_is_a_test_case_1, hypothesis, this_is_a_test_case(test_constant), inference(resolution, [], [one, two])).
    cnf(this_is_a_test_case_2, hypothesis, ~this_is_a_test_case(test_constant)).
    cnf(test_axiom, hypothesis, test_constant = X).
    cnf(test_axiom_2, hypothesis, ~test_constant = 0).
    """

    def __init__(self):
        # pylint: disable=unspecified-encoding
        self.parser = Lark(
            files("gym_saturation")
            .joinpath(os.path.join("resources", "TPTP.lark"))
            .read_text(),
            start="tptp_file",
        )

    def parse(self, tptp_text: str, tptp_folder: str) -> List[Clause]:
        """
        recursively parse a string containing a TPTP problem

        :param tptp_text: a name of a problem (or axioms) file
        :param tptp_folder: a folder containing TPTP database
        :returns: a list of clauses (including those of the axioms)
        """
        problem_tree = self.parser.parse(tptp_text)
        clauses = [
            CNFParser().transform(cnf_formula)
            for cnf_formula in problem_tree.find_data("cnf_annotated")
        ]
        for include in problem_tree.find_data("include"):
            token = include.children[0]
            if isinstance(token, Token):
                with open(
                    os.path.join(tptp_folder, token.value.replace("'", "")),
                    "r",
                    encoding="utf-8",
                ) as included_file:
                    clauses.extend(
                        self.parse(included_file.read(), tptp_folder)
                    )
        return clauses



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Grammar
********
"""
from dataclasses import dataclass, field
from typing import List, Optional, Union
from uuid import uuid1


@dataclass
class Variable:
    """
    .. _variable:

    a variable is characterised only by its name
    """

    name: str


@dataclass
class Function:
    """
    .. _Function:

    a functional symbol might be applied to a list of arguments
    """

    name: str
    arguments: List[Union[Variable, "Function"]]


Term = Union[Variable, Function]


@dataclass
class Predicate:
    """
    .. _Predicate:

    a predicate symbol might be applied to a list of arguments
    """

    name: str
    arguments: List[Term]


Proposition = Union[Predicate, Term]


@dataclass
class Literal:
    """
    .. _Literal:

    literal is an atom which can be negated or not
    """

    negated: bool
    atom: Predicate


def _term_to_tptp(term: Term) -> str:
    if isinstance(term, Function):
        arguments = [_term_to_tptp(argument) for argument in term.arguments]
        if arguments != []:
            return f"{term.name}({','.join(arguments)})"
    return term.name


def _literal_to_tptp(literal: Literal) -> str:
    res = "~" if literal.negated else ""
    if literal.atom.name != "=":
        res += (
            literal.atom.name
            + "("
            + ", ".join(
                [_term_to_tptp(term) for term in literal.atom.arguments]
            )
            + ")"
        )
    else:
        res += (
            _term_to_tptp(literal.atom.arguments[0])
            + " = "
            + _term_to_tptp(literal.atom.arguments[1])
        )
    return res


@dataclass
class Clause:
    """
    .. _Clause:

    clause is a disjunction of literals

    :param literals: a list of literals, forming the clause
    :param label: comes from the problem file or starts with ``inferred_`` if
         inferred during the episode
    :param inference_parents: a list of labels from which the clause was
         inferred. For clauses from the problem statement, this list is empty
    :param inference_rule: the rule according to which the clause was got from
         the ``inference_parents``
    :param processed: boolean value splitting clauses into unprocessed and
         processed ones; in the beginning, everything is not processed
    :param birth_step: a number of the step when the clause appeared in the
         unprocessed set; clauses from the problem have ``birth_step`` zero
    """

    literals: List[Literal]
    label: str = field(
        default_factory=lambda: "x" + str(uuid1()).replace("-", "_")
    )
    inference_parents: Optional[List[str]] = None
    inference_rule: Optional[str] = None
    processed: Optional[bool] = None
    birth_step: Optional[int] = None

    def __repr__(self):
        res = f"cnf({self.label}, hypothesis, "
        for literal in self.literals:
            res += _literal_to_tptp(literal) + " | "
        if res[-2:] == "| ":
            res = res[:-3]
        if not self.literals:
            res += "$false"
        if (
            self.inference_parents is not None
            and self.inference_rule is not None
        ):
            res += (
                f", inference({self.inference_rule}, [], ["
                + ", ".join(self.inference_parents)
                + "])"
            )
        return res + ")."



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Substitution
=============
"""
from dataclasses import dataclass
from typing import Union

from gym_saturation import grammar
from gym_saturation.utils import pickle_copy


@dataclass
class Substitution:
    """
    a mapping from ``Variable`` to ``Term``

    >>> substitution = Substitution(grammar.Variable("X"), grammar.Function("this_is_a_test_case", []))
    >>> substitution(grammar.Clause([grammar.Literal(False, grammar.Predicate("p", [grammar.Function("this_is_a_test_case", [grammar.Variable("X")])]))]))  # doctest: +ELLIPSIS
    cnf(..., hypothesis, p(this_is_a_test_case(this_is_a_test_case))).
    >>> substitution(grammar.Variable("X"))
    Function(name='this_is_a_test_case', arguments=[])
    """

    variable: grammar.Variable
    term: grammar.Term

    def __call__(
        self, target: Union[grammar.Clause, grammar.Proposition]
    ) -> Union[grammar.Clause, grammar.Proposition]:
        if isinstance(target, grammar.Clause):
            return self.substitute_in_clause(target)
        if isinstance(target, grammar.Predicate):
            return self._substitute_in_predicate(target)
        return self._substitute_in_term(target)

    def _substitute_in_term(self, term: grammar.Term) -> grammar.Term:
        if isinstance(term, grammar.Function):
            return grammar.Function(
                term.name,
                [
                    self._substitute_in_term(argument)
                    for argument in term.arguments
                ],
            )
        if term.name == self.variable.name:
            return pickle_copy(self.term)
        return pickle_copy(term)

    def _substitute_in_predicate(
        self, predicate: grammar.Predicate
    ) -> grammar.Predicate:
        return grammar.Predicate(
            predicate.name,
            [
                self._substitute_in_term(argument)
                for argument in predicate.arguments
            ],
        )

    def substitute_in_clause(self, clause: grammar.Clause) -> grammar.Clause:
        """
        apply the substitution to something which is known to be a ``Clause``

        :param clause: a clause to apply substitution to
        :returns: the result of the substitution
        """
        literals = []
        for literal in clause.literals:
            literals.append(
                grammar.Literal(
                    literal.negated,
                    self._substitute_in_predicate(literal.atom),
                )
            )
        new_clause: grammar.Clause = pickle_copy(clause)
        new_clause.literals = literals
        return new_clause



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Clause Space
=============
"""
from gym import spaces

from gym_saturation.grammar import Clause
from gym_saturation.parsing.json_grammar import dict_to_clause


class ClauseSpace(spaces.Space):
    """
    .. _clause_space:

    an OpenAI Gym space for a list of clauses

    >>> space = ClauseSpace()
    >>> space.sample() in space
    True
    >>> "no" in space
    False
    >>> [{"class": "no"}] in space
    False
    """

    def __init__(self):
        super().__init__(shape=(0,))

    def contains(self, x):
        if isinstance(x, list):
            for clause in x:
                try:
                    return isinstance(dict_to_clause(clause), Clause)
                except (ValueError, KeyError):
                    return False
        return False

    def sample(self):
        return [
            {
                "class": "Clause",
                "literals": [],
                "label": "test",
                "birth_step": 0,
                "processed": False,
                "inference_parents": [],
                "inference_rule": "",
            }
        ]



# Copyright 2021-2022 Boris Shminke

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     https://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Reflexivity Resolution
=======================
"""
from typing import List, Tuple

from gym_saturation import grammar, utils
from gym_saturation.logic_ops.unification import (
    NonUnifiableError,
    most_general_unifier,
)


def reflexivity_resolution(
    given_clause: grammar.Clause, a_literal: Tuple[grammar.Term, grammar.Term]
) -> grammar.Clause:
    r"""
    .. _reflexivity_resolution:

    reflexivity resolution rule

    .. math:: \frac{C\vee s\not\approx t}{\sigma\left(C\right)}

    where

    * :math:`C` and is a clause
    * :math:`s` and :math:`t` are terms, :math:`\not\approx` is a negation of equality
    * :math:`\sigma` is a most general unifier of :math:`s` and :math:`t`

    >>> from gym_saturation.grammar import Predicate, Variable, Function
    >>> reflexivity_resolution(grammar.Clause([grammar.Literal(True, Predicate("this_is_a_test_case", [Variable("X")]))]), [Variable("X"), Function("f", [])]).literals
    [Literal(negated=True, atom=Predicate(name='this_is_a_test_case', arguments=[Function(name='f', arguments=[])]))]

    :param given_clause: :math:`C`
    :param a_literal: :math:`s\not\approx t`
    :returns: a new clause --- the reflexivity resolution result
    """
    substitutions = most_general_unifier([a_literal[0], a_literal[1]])
    new_literals = utils.pickle_copy(given_clause.literals)
    result = grammar.Clause(new_literals)
    for substitution in substitutions:
        result = substitution.substitute_in_clause(result)
    return result


def all_possible_reflexivity_resolvents(
    given_clause: grammar.Clause,
) -> List[grammar.Clause]:
    """
    one of the four basic building blocks of the Given Clause algorithm

    >>> from gym_saturation.parsing.tptp_parser import TPTPParser
    >>> parser = TPTPParser()
    >>> clause = parser.parse("cnf(this_is_a_test_case, axiom, p(X) | ~ X=a | b != a).", "")[0]
    >>> all_possible_reflexivity_resolvents(clause)  # doctest: +ELLIPSIS
    [cnf(..., hypothesis, p(a) | ~b = a, inference(reflexivity_resolution, [], [this_is_a_test_case])).]

    :param given_clause: a new clause which should be combined with all the
        processed ones
    :returns: results of all possible reflexivity resolvents with each one from
        ``clauses`` and the ``given_clause``
    """
    reflexivity_resolvents: List[grammar.Clause] = []
    for i, a_literal in enumerate(given_clause.literals):
        if (
            a_literal.negated
            and a_literal.atom.name == "="
            or not a_literal.negated
            and a_literal.atom.name == "!="
        ) and len(a_literal.atom.arguments) == 2:
            a_clause = grammar.Clause(
                given_clause.literals[:i] + given_clause.literals[i + 1 :]
            )
            if a_clause.literals:
                try:
                    reflexivity_resolvents.append(
                        reflexivity_resolution(
                            a_clause,
                            (
                                a_literal.atom.arguments[0],
                                a_literal.atom.arguments[1],
                            ),
                        )
                    )
                except NonUnifiableError:
                    pass
    return [
        grammar.Clause(
            literals=reflexivity_resolvent.literals,
            inference_parents=[given_clause.label],
            inference_rule="reflexivity_resolution",
        )
        for ord_num, reflexivity_resolvent in enumerate(reflexivity_resolvents)
    ]
